/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.io;

import jalview.api.ComplexAlignFile;
import jalview.api.FeatureSettingsModelI;
import jalview.api.FeaturesDisplayedI;
import jalview.api.FeaturesSourceI;
import jalview.bin.Cache;
import jalview.bin.Jalview;
import jalview.datamodel.AlignmentI;
import jalview.datamodel.HiddenColumns;
import jalview.datamodel.PDBEntry;
import jalview.datamodel.SequenceI;
import jalview.gui.AlignFrame;
import jalview.gui.AlignViewport;
import jalview.gui.Desktop;
import jalview.gui.JvOptionPane;
import jalview.json.binding.biojson.v1.ColourSchemeMapper;
import jalview.project.Jalview2XML;
import jalview.schemes.ColourSchemeI;
import jalview.util.MessageManager;
import jalview.util.Platform;
import jalview.ws.utils.UrlDownloadClient;

import java.awt.Dimension;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import javax.swing.SwingUtilities;

public class FileLoader implements Runnable
{
  private static final String DEFAULT_FILE_FORMAT_PROPERTY = "DEFAULT_FILE_FORMAT";

  private static final String RECENT_URL_PROPERTY = "RECENT_URL";

  private static final String RECENT_FILE_PROPERTY = "RECENT_FILE";

  private static final String TAB = "\t";

  /*
   * maximum number of items in file/url history lists;
   * pseudo-constant (not final) so amendable by Groovy if wanted
   */
  private static int MAX_HISTORY = 11;

  String file;

  DataSourceType protocol;

  FileFormatI format;

  AlignmentFileReaderI source; // alternative specification of where data
                               // comes from

  AlignViewport viewport;

  AlignFrame alignFrame;

  long loadtime;

  long memused;

  boolean raiseGUI = true;

  private File selectedFile;

  /**
   * default constructor always raised errors in GUI dialog boxes
   */
  public FileLoader()
  {
    this(true);
  }

  /**
   * construct a Fileloader that may raise errors non-interactively
   * 
   * @param raiseGUI
   *          true if errors are to be raised as GUI dialog boxes
   */
  public FileLoader(boolean raiseGUI)
  {
    this.raiseGUI = raiseGUI;
  }

  /**
   * It is critical that all these fields are set, as this instance is reused.
   * 
   * @param source
   * @param file
   * @param inFile
   * @param dataSourceType
   * @param format
   */
  private void setFileFields(AlignmentFileReaderI source, File file,
          String inFile, DataSourceType dataSourceType, FileFormatI format)
  {
    this.source = source;
    this.file = inFile;
    this.selectedFile = file;
    this.protocol = dataSourceType;
    this.format = format;
  }


  /**
   * Uppercase LoadFile is deprecated because it does not pass byte[] data to
   * JavaScript
   * 
   * @param viewport
   * @param file
   * @param protocol
   * @param format
   */
  @Deprecated
  public void LoadFile(AlignViewport viewport, String file,
          DataSourceType protocol, FileFormatI format)
  {
    if (viewport != null)
    {
      this.viewport = viewport;
    }
    loadFile(file, protocol, format);
  }

  public void LoadFile(AlignViewport viewport, File file,
          DataSourceType protocol, FileFormatI format)
  {
    loadFile(viewport, file, protocol, format);
  }

  /**
   * Uppercase LoadFile is deprecated because it does not pass byte[] data to
   * JavaScript
   * 
   * @param file
   * @param protocol
   * @param format
   */
  @Deprecated
  public void LoadFile(String file, DataSourceType protocol,
          FileFormatI format)
  {
    loadFile(file, protocol, format);
  }

  /**
   * necessary to use Object here in order to pass the file data
   * 
   * @param viewport
   * @param file
   *          File preferably to String
   * @param protocol
   * @param format
   */
  public void loadFile(AlignViewport viewport, File file,
          DataSourceType protocol, FileFormatI format)
  {
    if (viewport != null)
    {
      this.viewport = viewport;
    }
    this.selectedFile = file;
    loadFile(selectedFile.getPath(), protocol, format);
  }

  private void loadFile(String file, DataSourceType protocol,
          FileFormatI format)
  {
    this.file = file;
    this.protocol = protocol;
    this.format = format;

    final Thread loader = new Thread(this);

    SwingUtilities.invokeLater(new Runnable()
    {
      @Override
      public void run()
      {
        loader.start();
      }
    });
  }

  /**
   * Load alignment from (file, protocol) and wait till loaded
   * 
   * @param file
   * @param sourceType
   * @return alignFrame constructed from file contents
   */
  @Deprecated
  public AlignFrame LoadFileWaitTillLoaded(String file,
          DataSourceType sourceType)
  {
    return LoadFileWaitTillLoaded(file, sourceType, null);
  }

  /**
   * Load alignment from (file, protocol) of type format and wait till loaded
   * 
   * @param file
   * @param sourceType
   * @param format
   * @return alignFrame constructed from file contents
   */
  @Deprecated
  public AlignFrame LoadFileWaitTillLoaded(String file,
          DataSourceType sourceType, FileFormatI format)
  {
    setFileFields(null, null, file, sourceType, format);
    return _loadFileWaitTillLoaded();
  }

  /**
   * Load alignment from (file, protocol) of type format and wait till loaded
   * 
   * @param fileObject
   * @param sourceType
   * @param format
   * @return alignFrame constructed from file contents
   * @throws NullPointerException
   *           if {@code file} is null
   */
  public AlignFrame loadFileWaitTillLoaded(File fileObject,
          DataSourceType sourceType, FileFormatI format)
  {
    setFileFields(null, fileObject, fileObject.getPath(), sourceType, format);
    return _loadFileWaitTillLoaded();
  }

  /**
   * Runs the 'run' method (in this thread), then returns the alignFrame that
   * (hopefully) has been constructed
   * 
   * @return
   */
  private AlignFrame _loadFileWaitTillLoaded()
  {
    this.run();
    return alignFrame;
  }

  /**
   * Updates the Jalview properties file to add the last file or URL read to the
   * tab-separated list in property RECENT_FILE or RECENT_URL respectively. The
   * property is created if it does not already exist. The new entry is added as
   * the first item (without duplication). If the list exceeds some maximum
   * length (currently 10 entries), the oldest (last) entry is dropped.
   * <p>
   * Files read from the temporary file directory are not added to the property.
   * <p>
   * Property DEFAULT_FILE_FORMAT (e.g. "Fasta" etc) is also set if currently
   * reading from file.
   */
  public void updateRecentlyOpened()
  {
    List<String> recent = new ArrayList<>();
    if (protocol == DataSourceType.PASTE)
    {
      // do nothing if the file was pasted in as text... there is no filename to
      // refer to it as.
      return;
    }
    // BH logic change here just includes ignoring file==null
    if (file == null
            || file.indexOf(System.getProperty("java.io.tmpdir")) > -1)
    {
      // ignore files loaded from the system's temporary directory
      return;
    }
    String type = protocol == DataSourceType.FILE ? RECENT_FILE_PROPERTY
            : RECENT_URL_PROPERTY;

    String historyItems = Cache.getProperty(type);

    // BH simpler coding
    if (historyItems != null)
    {
      String[] tokens = historyItems.split("\\t");

      for (String token : tokens)
      {
        recent.add(token.trim());
      }
    }

    if (recent.contains(file))
    {
      recent.remove(file);
    }

    StringBuilder newHistory = new StringBuilder(file);
    for (int i = 0; i < recent.size() && i < MAX_HISTORY - 1; i++)
    {
      newHistory.append(TAB);
      newHistory.append(recent.get(i));
    }

    Cache.setProperty(type, newHistory.toString());

    if (protocol == DataSourceType.FILE)
    {
      Cache.setProperty(DEFAULT_FILE_FORMAT_PROPERTY, format.getName());
    }
  }

  @Override
  public void run()
  {
    String title = protocol == DataSourceType.PASTE
            ? "Copied From Clipboard"
            : file;
    Runtime rt = Runtime.getRuntime();
    try
    {
      if (Desktop.getInstance() != null)
      {
        Desktop.getInstance().startLoading(file);
      }
      if (format == null)
      {
        // just in case the caller didn't identify the file for us
        if (source != null)
        {
          format = new IdentifyFile().identify(source, false);
          // identify stream and rewind rather than close
        }
        else if (selectedFile != null) {
          format = new IdentifyFile().identify(selectedFile, protocol);
        }
        else
        {
          format = new IdentifyFile().identify(file, protocol);
        }

      }

      if (format == null)
      {
        Desktop.getInstance().stopLoading();
        System.err.println("The input file \"" + file
                + "\" has null or unidentifiable data content!");
        if (!Jalview.isHeadlessMode())
        {
          JvOptionPane.showInternalMessageDialog(Desktop.getDesktopPane(),
                  MessageManager.getString("label.couldnt_read_data")
                          + " in " + file + "\n"
                          + AppletFormatAdapter.getSupportedFormats(),
                  MessageManager.getString("label.couldnt_read_data"),
                  JvOptionPane.WARNING_MESSAGE);
        }
        return;
      }
      // TODO: cache any stream datasources as a temporary file (eg. PDBs
      // retrieved via URL)
      if (Desktop.getDesktopPane() != null && Desktop.getDesktopPane().isShowMemoryUsage())
      {
        System.gc();
        memused = (rt.maxMemory() - rt.totalMemory() + rt.freeMemory()); // free
        // memory
        // before
        // load
      }
      loadtime = -System.currentTimeMillis();
      AlignmentI al = null;

      if (FileFormat.Jalview.equals(format))
      {
        if (source != null)
        {
          // Tell the user (developer?) that this is going to cause a problem
          System.err.println(
                  "IMPLEMENTATION ERROR: Cannot read consecutive Jalview XML projects from a stream.");
          // We read the data anyway - it might make sense.
        }
        // BH 2018 switch to File object here instead of filename
        Platform.timeCheck(null, Platform.TIME_MARK);
        alignFrame = new Jalview2XML(raiseGUI).loadJalviewAlign(selectedFile == null ? file : selectedFile);
        Platform.timeCheck("JVP loaded", Platform.TIME_MARK);

      }
      else
      {
        String error = AppletFormatAdapter.getSupportedFormats();
        try
        {
          if (source != null)
          {
            // read from the provided source
            al = new FormatAdapter().readFromFile(source, format);
          }
          else
          {

            // open a new source and read from it
            FormatAdapter fa = new FormatAdapter();
            boolean downloadStructureFile = format.isStructureFile()
                    && protocol.equals(DataSourceType.URL);
            if (downloadStructureFile)
            {
              String structExt = format.getExtensions().split(",")[0];
              int pt = file.lastIndexOf(file.indexOf('/') >= 0 ? "/"
                      : System.getProperty("file.separator"));
              String urlLeafName = file.substring(pt,
                      file.lastIndexOf("."));
              String tempStructureFileStr = createNamedJvTempFile(
                      urlLeafName, structExt);
              
              // BH - switching to File object here so as to hold on to the
              // bytes array directly
              File tempFile = new File(tempStructureFileStr);
              UrlDownloadClient.download(file, tempFile);
              
              al = fa.readFile(tempFile, DataSourceType.FILE,
                      format);
              source = fa.getAlignFile();
            }
            else
            {
              if (selectedFile == null) {
                al = fa.readFile(file, protocol, format);
                
              } else {
                al = fa.readFile(selectedFile, protocol, format);
                             }
              source = fa.getAlignFile(); // keep reference for later if
              
                                          // necessary.
            }
          }
        } catch (java.io.IOException ex)
        {
          error = ex.getMessage();
        }

        if ((al != null) && (al.getHeight() > 0) && al.hasValidSequence())
        {
          // construct and register dataset sequences
          for (SequenceI sq : al.getSequences())
          {
            while (sq.getDatasetSequence() != null)
            {
              sq = sq.getDatasetSequence();
            }
            if (sq.getAllPDBEntries() != null)
            {
              for (PDBEntry pdbe : sq.getAllPDBEntries())
              {
                // register PDB entries with desktop's structure selection
                // manager
                Desktop.getStructureSelectionManager()
                        .registerPDBEntry(pdbe);
              }
            }
          }

          FeatureSettingsModelI proxyColourScheme = source
                  .getFeatureColourScheme();
          if (viewport != null)
          {
            if (proxyColourScheme != null)
            {
              viewport.applyFeaturesStyle(proxyColourScheme);
            }
            // append to existing alignment
            viewport.addAlignment(al, title);
          }
          else
          {
            // otherwise construct the alignFrame

            if (source instanceof ComplexAlignFile)
            {
              HiddenColumns colSel = ((ComplexAlignFile) source)
                      .getHiddenColumns();
              SequenceI[] hiddenSeqs = ((ComplexAlignFile) source)
                      .getHiddenSequences();
              String colourSchemeName = ((ComplexAlignFile) source)
                      .getGlobalColourScheme();
              FeaturesDisplayedI fd = ((ComplexAlignFile) source)
                      .getDisplayedFeatures();
              alignFrame = new AlignFrame(al, hiddenSeqs, colSel,
                      AlignFrame.DEFAULT_WIDTH, AlignFrame.DEFAULT_HEIGHT);
              alignFrame.getViewport().setFeaturesDisplayed(fd);
              alignFrame.getViewport().setShowSequenceFeatures(
                      ((ComplexAlignFile) source).isShowSeqFeatures());
              ColourSchemeI cs = ColourSchemeMapper
                      .getJalviewColourScheme(colourSchemeName, al);
              if (cs != null)
              {
                alignFrame.changeColour(cs);
              }
            }
            else
            {
              alignFrame = new AlignFrame(al, AlignFrame.DEFAULT_WIDTH,
                      AlignFrame.DEFAULT_HEIGHT);
              if (source instanceof FeaturesSourceI)
              {
                alignFrame.getViewport().setShowSequenceFeatures(true);
              }
            }
            // add metadata and update ui
            if (!(protocol == DataSourceType.PASTE))
            {
              alignFrame.setFileName(file, format);
              alignFrame.setFileObject(selectedFile); // BH 2018 SwingJS
            }
            if (proxyColourScheme != null)
            {
              alignFrame.getViewport()
                      .applyFeaturesStyle(proxyColourScheme);
            }
            alignFrame.setStatus(MessageManager.formatMessage(
                    "label.successfully_loaded_file", new String[]
                    { title }));

            if (raiseGUI)
            {
              // add the window to the GUI
              // note - this actually should happen regardless of raiseGUI
              // status in Jalview 3
              // TODO: define 'virtual desktop' for benefit of headless scripts
              // that perform queries to find the 'current working alignment'
              
              
              Dimension dim = Platform.getDimIfEmbedded(alignFrame,
                      AlignFrame.DEFAULT_WIDTH, AlignFrame.DEFAULT_HEIGHT);
              alignFrame.setSize(dim);
              Desktop.addInternalFrame(alignFrame, title, dim.width,
                      dim.height);
            }

            try
            {
              alignFrame.setMaximum(jalview.bin.Cache
                      .getDefault("SHOW_FULLSCREEN", false));
            } catch (java.beans.PropertyVetoException ex)
            {
            }
          }
        }
        else
        {
          if (Desktop.getInstance() != null)
          {
            Desktop.getInstance().stopLoading();
          }

          final String errorMessage = MessageManager.getString(
                  "label.couldnt_load_file") + " " + title + "\n" + error;
          // TODO: refactor FileLoader to be independent of Desktop / Applet GUI
          // bits ?
          if (raiseGUI && Desktop.getDesktopPane() != null)
          {
            javax.swing.SwingUtilities.invokeLater(new Runnable()
            {
              @Override
              public void run()
              {
                JvOptionPane.showInternalMessageDialog(Desktop.getDesktopPane(),
                        errorMessage,
                        MessageManager
                                .getString("label.error_loading_file"),
                        JvOptionPane.WARNING_MESSAGE);
              }
            });
          }
          else
          {
            System.err.println(errorMessage);
          }
        }
      }

      updateRecentlyOpened();

    } catch (Exception er)
    {
      System.err.println("Exception whilst opening file '" + file);
      er.printStackTrace();
      if (raiseGUI)
      {
        javax.swing.SwingUtilities.invokeLater(new Runnable()
        {
          @Override
          public void run()
          {
            JvOptionPane.showInternalMessageDialog(Desktop.getDesktopPane(),
                    MessageManager.formatMessage(
                            "label.problems_opening_file", new String[]
                            { file }),
                    MessageManager.getString("label.file_open_error"),
                    JvOptionPane.WARNING_MESSAGE);
          }
        });
      }
      alignFrame = null;
    } catch (OutOfMemoryError er)
    {

      er.printStackTrace();
      alignFrame = null;
      if (raiseGUI)
      {
        javax.swing.SwingUtilities.invokeLater(new Runnable()
        {
          @Override
          public void run()
          {
            JvOptionPane.showInternalMessageDialog(Desktop.getDesktopPane(),
                    MessageManager.formatMessage(
                            "warn.out_of_memory_loading_file", new String[]
                            { file }),
                    MessageManager.getString("label.out_of_memory"),
                    JvOptionPane.WARNING_MESSAGE);
          }
        });
      }
      System.err.println("Out of memory loading file " + file + "!!");

    }
    loadtime += System.currentTimeMillis();
    // TODO: Estimate percentage of memory used by a newly loaded alignment -
    // warn if more memory will be needed to work with it
    // System.gc();
    memused = memused
            - (rt.maxMemory() - rt.totalMemory() + rt.freeMemory()); // difference
    // in free
    // memory
    // after
    // load
    if (Desktop.getDesktopPane() != null && Desktop.getDesktopPane().isShowMemoryUsage())
    {
      if (alignFrame != null)
      {
        AlignmentI al = alignFrame.getViewport().getAlignment();

        System.out.println("Loaded '" + title + "' in "
                + (loadtime / 1000.0) + "s, took an additional "
                + (1.0 * memused / (1024.0 * 1024.0)) + " MB ("
                + al.getHeight() + " seqs by " + al.getWidth() + " cols)");
      }
      else
      {
        // report that we didn't load anything probably due to an out of memory
        // error
        System.out.println("Failed to load '" + title + "' in "
                + (loadtime / 1000.0) + "s, took an additional "
                + (1.0 * memused / (1024.0 * 1024.0))
                + " MB (alignment is null)");
      }
    }
    // remove the visual delay indicator
    if (Desktop.getInstance() != null)
    {
      Desktop.getInstance().stopLoading();
    }

  }

  /**
   * This method creates the file -
   * {tmpdir}/jalview/{current_timestamp}/fileName.exetnsion using the supplied
   * file name and extension
   * 
   * @param fileName
   *          the name of the temp file to be created
   * @param extension
   *          the extension of the temp file to be created
   * @return
   */
  private static String createNamedJvTempFile(String fileName,
          String extension) throws IOException
  {
    String seprator = System.getProperty("file.separator");
    String jvTempDir = System.getProperty("java.io.tmpdir") + "jalview"
            + seprator + System.currentTimeMillis();
    File tempStructFile = new File(
            jvTempDir + seprator + fileName + "." + extension);
    tempStructFile.mkdirs();
    return tempStructFile.toString();
  }

  /**
   * 
   * @param file a File, or a String which is a name of a file
   * @return
   * @throws FileNotFoundException 
   */
  public static BufferedReader getBufferedReader(Object file) throws FileNotFoundException {
    if (file instanceof String)
    {
      return new BufferedReader(new FileReader((String) file));
    }
    byte[] bytes = Platform.getFileBytes((File) file);
    if (bytes != null)
    {
      return new BufferedReader(new InputStreamReader(new ByteArrayInputStream(bytes)));
    }
    return  new BufferedReader(new FileReader((File) file));
  }

}
